/** Program of compass creation
  * Using a GY83 circuit including 
  * a magnetometer and an accelerometer a temerature sensor and a pressure sensor
  Author : yves cordier

  Note this program uses units from 
  */

// *****************************************
// include zone
// *****************************************

#include <Wire.h>
#include <MechaQMC5883.h> // librarie from https://drive.google.com/file/d/1qQyX1cCCijx1q1IKiPZuhHgtWCmFrEmP/edit
#include <MPU6050.h> // librarie from https://github.com/jrowberg/i2cdevlib/tree/master/Arduino
#include <BMP180I2C.h> // librarie from https://bitbucket.org/christandlg/bmp180mi/src/master/

// #include <I2Cdev.h> // called automaticaly by HMC5883L and MPU6050 libraries but must be present on your librarie folder


// *****************************************
// define zone
// *****************************************

#define _delaiScrutation ((1000 / 15) + 1)  // chips where configured to 15 hz sampling maxi so we should wait a bit between two sampling
#define _magnetoAddress (0x0D)              // address on real chip 0x1E. on chineses clone it is 0x0D
#define _tempAddress (0x77)                 // address of temperature and pression sensor
#define _deltaAxes (-1.)                    // negative correction due to z axis positions (1. if directed down -1.if up)

#define horizontalLed (13)  // a led to indicate horizontality
#define _calibration (2)  // The calibration button

#define epsilon (1E-4)      // epsilon of gz to compute

// *****************************************
// complex data definition
// *****************************************

struct vecteurInt {
  int x;
  int y;
  int z;
};

struct vecteurFloat {
  float x;
  float y;
  float z;
};

// *****************************************
// global variables zone
// *****************************************
MPU6050 accelero;
MechaQMC5883 magneto;
BMP180I2C tempPressure(_tempAddress);


vecteurFloat offsetB;             // the offsets of magnetometer
vecteurInt maxiB, miniB;          // maxi and mini values of magnetometer
boolean initMagneto = true;       //
boolean calibrationDone = false;  // does the magnetometer calibration has been done ?

// *****************************************
// function zone
// *****************************************
/* setup  +--------
 * loop   +---- magnetoCalibration() -+-- setMaxiMiniB
 *        |                           +-- showMagnetoCAlibration
 *        |                                   
 *        |
 *        +---- getCap ---------------+--- spherizeValues;
 *        |                           +--- correctCap;
 *        *---- showCap
 *        +---- showAccelerometerTemperature
 *        +---- showExternalPressure
 *        +---- showExternalTemperature 
*/
// *****************************************
/** compute magnetometers values
 * @param x int value on x axis of magnetic fielg (according to resolution of chip)
 * @param y int value on y axis of magnetic fielg (according to resolution of chip)
 * @param z int value on z axis of magnetic fielg (according to resolution of chip)
 * @param first boolean value which indicate the reset of process 
 */
void setMaxiMiniB(int x, int y, int z, boolean first) {
  if (first) {
    maxiB.x = x;
    miniB.x = x;
    maxiB.y = y;
    miniB.y = y;
    maxiB.z = z;
    miniB.z = z;
  }
  if (x > maxiB.x) maxiB.x = x;
  if (x < miniB.x) miniB.x = x;
  if (y > maxiB.y) maxiB.y = y;
  if (y < miniB.y) miniB.y = y;
  if (z > maxiB.z) maxiB.z = z;
  if (z < miniB.z) miniB.z = z;
  offsetB.x = (maxiB.x + miniB.x) / 2;
  offsetB.y = (maxiB.y + miniB.y) / 2;
  offsetB.z = (maxiB.z + miniB.z) / 2;
}

/** Showing stored calibration values on Serial beam
 */
void showMagnetoCalibration() {
  Serial.println();
  Serial.print("results of calibration :");
  Serial.println("**********************************************************************************");
  Serial.print("Bxmin : ");
  Serial.print(miniB.x);
  Serial.print(" Bxmax : ");
  Serial.print(maxiB.x);
  Serial.print(" Xoffset : ");
  Serial.println(offsetB.x);
  Serial.print("Bymin : ");
  Serial.print(miniB.y);
  Serial.print(" Bymax : ");
  Serial.print(maxiB.y);
  Serial.print(" Yoffset : ");
  Serial.println(offsetB.y);
  Serial.print("Bzmin : ");
  Serial.print(miniB.z);
  Serial.print(" Bzmax : ");
  Serial.print(maxiB.z);
  Serial.print(" Zoffset : ");
  Serial.println(offsetB.z);
  Serial.println("**********************************************************************************");
}

/** Making magnetometer calibration
 * according to calibrationDone variable which indicate if a calibration has to be made
 * if so, calibration is lauched and acquire of values is done
 * while operator moving sensor in all directions
 * @param maxi number of points of calibration (500 if ommited)
  */
void magnetoCalibration(int maxi = 500) {
  if (!calibrationDone) {
    int bx, by, bz;
    magneto.read(&bx, &by, &bz);
    setMaxiMiniB(bx, by, bz, true);  // true : first setting to initiate
    delay(_delaiScrutation);
    Serial.println("magnetometer calibration in progress");
    Serial.println(" press any key to go on ..");
    Serial.flush();                             // clear input buffer
    while (Serial.peek() == -1) { delay(10); }  // wait until keypressed
    Serial.flush();                             // clear buffer
    for (int i = 0; i < maxi; i++) {
      magneto.read(&bx, &by, &bz);
      setMaxiMiniB(bx, by, bz, false);  // false : do not initiate process
      if (i % 50 == 0) {
        Serial.print("  ");
        Serial.print(int((float(i) / maxi) * 100));
        Serial.print('%');
        Serial.println(" done");
      }
      Serial.print('.');
      delay(_delaiScrutation);  // set to 15 Hz sampling rate so could not ask more quicquely
    }
    showMagnetoCalibration();
    calibrationDone = true;
  }
}

void scanI2C() {
  byte error, address;
  int nDevices;

  Serial.println("Scanning...");

  nDevices = 0;
  for (address = 8; address < 127; address++) {
    Wire.beginTransmission(address);
    error = Wire.endTransmission();

    if (error == 0) {
      Serial.print("I2C device found at address 0x");
      if (address < 16)
        Serial.print("0");
      Serial.print(address, HEX);
      Serial.println(" !");

      nDevices++;
    } else if (error == 4) {
      Serial.print("Unknow error at address 0x");
      if (address < 16)
        Serial.print("0");
      Serial.println(address, HEX);
    }
  }
  if (nDevices == 0)
    Serial.println("No I2C devices found\n");
  else
    Serial.println("done\n");

  delay(2000);
}

/** initialisation function
*/
void setup() {
  magneto.setAddress(_magnetoAddress);  // as my magnetometer is a chiness clone, I have to indicate the real adress (0x0D)

  Serial.begin(9600);  // to discuss with IT console
  Wire.begin();        // initiate the I2C protocol

  // initialize devices
  Serial.println(" The magnetometer should not apears at the first power on (0x0d or 0x1E)");
  scanI2C();  // result of scan (0x68 : accelerometer; 0x77 : pression meter)
  Serial.println("Initializing I2C devices...");
  // initialize mpu6050
  accelero.initialize();
  accelero.setI2CBypassEnabled(true);  // set bypass mode for gateway to hmc5883L
  accelero.setI2CMasterModeEnabled(false);

  // from this point we could access to magnetometer by I2C bus through accelerometer bus extension
  Serial.println(" The magnetometer has to apears (0x0d or 0x1E)");
  scanI2C();  // result of scan (0x0D, 0x68, 0x77)

  magneto.init();  // set sampling mode and so on
  int vx, vy, vz;

  // first attempt to get values from magnetometer
  magneto.read(&vx, &vy, &vz);  // first dummy acces by safety

  if (!tempPressure.begin()) {
    Serial.println("begin() failed. check your BMP180 Interface and I2C Address.");
    while (1)
      ;
  }
  //reset sensor to default parameters.
  tempPressure.resetToDefaults();
  //enable ultra high resolution mode for pressure measurements
  tempPressure.setSamplingMode(BMP180MI::MODE_UHR);

  // the horizontality led
  pinMode(horizontalLed, OUTPUT);
  digitalWrite(horizontalLed, false);
  // the calibration button
  pinMode(_calibration, INPUT_PULLUP);
}

/** correction of values read from magnetometer
 * according corrections dues to non verticality of z axis
 * detected by accelerometer
 * @param bx value return by MP6050 on x axis
 * @param by value return by MP6050 on y axis
 * @param bz value return by MP6050 on z axis
*/
float correctCap(float bx, float by, float bz) {
  float bxcorrige = bx;
  float bycorrige = by;
  int gx, gy, gz;

  accelero.getAcceleration(&gx, &gy, &gz);  // read from accelerometer
  float gxf = gx;
  float gyf = gy;
  float gzf = gz;
  /*Serial.print(gxf);
  Serial.print("\t");
  Serial.print(gyf);
  Serial.print("\t");
  Serial.print(gzf);
  Serial.print("\t");
  */
  // compute corrections
  if (abs(gzf) > epsilon) {
    bxcorrige = bx - (gxf / gzf) * bz;
    bycorrige = _deltaAxes * (by + (gyf / gzf) * bz);

    // compute angle
    float a = atan2(abs(bycorrige), abs(bxcorrige)) * 180 / M_PI;
    // normalize angle
    if (bxcorrige <= 0) {
      if (bycorrige <= 0) {
        a = 180 - a;
      } else {
        a = a + 180;
      }
    } else {
      if (bycorrige <= 0) {  // ne rien faire
      } else {
        a = 360 - a;
      };
    }

    while (a >= 360) a -= 360;
    while (a < 0) a += 360;
    // indicate horizontality or nearly
    digitalWrite(horizontalLed, (abs(gxf / gzf) < 0.05) && (abs(gyf / gzf) < 0.05));
    /*Serial.print(abs(gxf / gzf));
    Serial.print("\t");
    Serial.print(abs(gyf / gzf));
    Serial.print("\t");
    Serial.print((abs(gxf / gzf) < 0.05) && (abs(gyf / gzf) < 0.05));
    */
    return a;
  } else {
    digitalWrite(horizontalLed, false);
    return 0;
  }
}
/** correction of values read from magnetometer
 * according offsets and proportional values computed on calibration
 * detected by accelerometer
 * @param bx value return by MP6050 on x axis (float)
 * @param by value return by MP6050 on y axis (float)
 * @param bz value return by MP6050 on z axis (float)
*/
void spherizeValues(float &bx, float &by, float &bz) {
  bx = (bx - offsetB.x) / (maxiB.x - miniB.x);
  by = (by - offsetB.y) / (maxiB.y - miniB.y);
  bz = (bz - offsetB.z) / (maxiB.z - miniB.z);
}

/** geting cap from magnetometer and tinker it
 * @return cap in degrs corrected making attention of accelerometer
 */
float getCap() {
  int bx, by, bz;
  float bxf, byf, bzf;
  magneto.read(&bx, &by, &bz);  // rough values
  bxf = bx;
  byf = by;
  bzf = bz;

  spherizeValues(bxf, byf, bzf);  // correct offsets and amplification on each axe regarding of calibration values

  return correctCap(bxf, byf, bzf);  // compute values getting regard of non horizontality of compass
}

/** Showing result of opération
 * this function could be change to make a visual one on graphic pseudo-compass
 * @param angle value of direction in degres (float)
*/
void showCap(float angle) {
  Serial.print("cap :");
  Serial.println(angle);
}

/** Showing internal temperature of 
 * this function could be change to make a visual one on graphic thermometer
 */
void showAccelerometerTemperature() {
  int temp = accelero.getTemperature();
  float tempf = (float)temp / 340. + 36.53;  // according to specifications
  Serial.print("temp : ");
  Serial.print(tempf);
  Serial.print("\t");
}

/** Showing external sensor temperature  
 * this function could be change to make a visual one on graphic thermometer
 * @return true if suucced false if not
 */
boolean showExternalTemperature() {
  //start a temperature measurement
  if (!tempPressure.measureTemperature()) {
    return false;
  }
  //wait for the measurement to finish. proceed as soon as hasValue() returned true.
  do {
    delay(10);
  } while (!tempPressure.hasValue());
  Serial.print("external temp : ");
  Serial.print(tempPressure.getTemperature());
  Serial.print(" °C \t");
  return true;
}
/** Showing external pressure   
 * this function could be change to make a visual one on graphic thermometer
 * @return true if suucced false if not
 */
boolean showExternalPressure() {

  //start a pressure measurement. pressure measurements depend on temperature measurement, you should only start a pressure
  //measurement immediately after a temperature measurement.
  if (!tempPressure.measurePressure()) {
    return false;
  }
  //wait for the measurement to finish. proceed as soon as hasValue() returned true.
  do {
    delay(10);
  } while (!tempPressure.hasValue());
  Serial.print("Pressure: ");
  Serial.print(tempPressure.getPressure());
  Serial.print(" Pa \t");
  return true;
}

/** main loop
*/
void loop() {
  // calibration on demand treatment
  static boolean etatPressButtonSto = false;
  boolean etatPressButton = !digitalRead(_calibration);  // when pressed _calibration pin is grounded
  if (etatPressButton & !etatPressButtonSto) calibrationDone = false;
  etatPressButtonSto = etatPressButton;

  magnetoCalibration();  // calibration if necessary ie at initiate and if asked by press button

  delay(_delaiScrutation);         // waiting according to max frequency of reading of magnetometer
  showAccelerometerTemperature();  // show internal temperature of accelerometer
  while (!showExternalTemperature()) {};       // show temperature of external sensor
  while (!showExternalPressure()) {};       // show pressure of external sensor
  float angle = getCap();          // The real job of this programm
  showCap(angle);                  // to share the result
}
